from typing import Optional

from pydantic import AnyUrl, BaseModel, Extra


class Contact(BaseModel):
    """
    Contact information for the exposed API.
    """

    name: Optional[str] = None
    """
    The identifying name of the contact person/organization.
    """

    url: Optional[AnyUrl] = None
    """
    The URL pointing to the contact information.
    MUST be in the form of a URL.
    """

    email: Optional[str] = None
    """
    The email address of the contact person/organization.
    MUST be in the form of an email address.
    """

    class Config:
        extra = Extra.ignore
        schema_extra = {
            "examples": [
                {"name": "API Support", "url": "http://www.example.com/support", "email": "support@example.com"}
            ]
        }



from typing import Dict

from .path_item import PathItem


Paths = Dict[str, PathItem]
"""
Holds the relative paths to the individual endpoints and their operations.
The path is appended to the URL from the [`Server Object`](#serverObject) in order to construct the full URL.

The Paths MAY be empty, due to [Access Control List (ACL) constraints](#securityFiltering).
"""

"""Patterned Fields"""

# "/{path}" : PathItem
"""
A relative path to an individual endpoint.
The field name MUST begin with a forward slash (`/`).
The path is **appended** (no relative URL resolution) to the expanded URL 
from the [`Server Object`](#serverObject)'s `url` field in order to construct the full URL.
[Path templating](#pathTemplating) is allowed.
When matching URLs, concrete (non-templated) paths would be matched before their templated counterparts.
Templated paths with the same hierarchy but different templated names MUST NOT exist as they are identical.
In case of ambiguous matching, it's up to the tooling to decide which one to use.
"""



from typing import Dict, Optional, Union

from pydantic import AnyUrl, BaseModel, Extra


class OAuthFlow(BaseModel):
    """
    Configuration details for a supported OAuth Flow
    """

    authorizationUrl: Optional[Union[AnyUrl, str]] = None
    """
    **REQUIRED** for `oauth2 ("implicit", "authorizationCode")`.
    The authorization URL to be used for this flow.
    This MUST be in the form of a URL.
    The OAuth2 standard requires the use of TLS.
    """

    tokenUrl: Optional[Union[AnyUrl, str]] = None
    """
    **REQUIRED** for `oauth2 ("password", "clientCredentials", "authorizationCode")`.
    The token URL to be used for this flow.
    This MUST be in the form of a URL.
    The OAuth2 standard requires the use of TLS.
    """

    refreshUrl: Optional[Union[AnyUrl, str]] = None
    """
    The URL to be used for obtaining refresh tokens.
    This MUST be in the form of a URL.
    The OAuth2 standard requires the use of TLS.
    """

    scopes: Optional[Dict[str, str]] = None
    """
    **REQUIRED** for `oauth2`. The available scopes for the OAuth2 security scheme.
    A map between the scope name and a short description for it.
    The map MAY be empty.
    """

    class Config:
        extra = Extra.ignore
        schema_extra = {
            "examples": [
                {
                    "authorizationUrl": "https://example.com/api/oauth/dialog",
                    "scopes": {"write:pets": "modify pets in your account", "read:pets": "read your pets"},
                },
                {
                    "authorizationUrl": "https://example.com/api/oauth/dialog",
                    "tokenUrl": "https://example.com/api/oauth/token",
                    "scopes": {"write:pets": "modify pets in your account", "read:pets": "read your pets"},
                },
                {
                    "authorizationUrl": "/api/oauth/dialog",  # issue #5: allow relative path
                    "tokenUrl": "/api/oauth/token",  # issue #5: allow relative path
                    "refreshUrl": "/api/oauth/token",  # issue #5: allow relative path
                    "scopes": {"write:pets": "modify pets in your account", "read:pets": "read your pets"},
                },
            ]
        }



from typing import List, Optional

from pydantic import BaseModel, Extra


class ServerVariable(BaseModel):
    """An object representing a Server Variable for server URL template substitution."""

    enum: Optional[List[str]] = None
    """
    An enumeration of string values to be used if the substitution options are from a limited set.
    The array SHOULD NOT be empty.
    """

    default: str = ...
    """
    **REQUIRED**. The default value to use for substitution,
    which SHALL be sent if an alternate value is _not_ supplied.
    Note this behavior is different than the [Schema Object's](#schemaObject) treatment of default values, 
    because in those cases parameter values are optional.
    If the [`enum`](#serverVariableEnum) is defined, the value SHOULD exist in the enum's values.
    """

    description: Optional[str] = None
    """
    An optional description for the server variable.
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text representation.
    """

    class Config:
        extra = Extra.ignore



from typing import Dict, List


SecurityRequirement = Dict[str, List[str]]
"""
Lists the required security schemes to execute this operation.
The name used for each property MUST correspond to a security scheme declared in the
[Security Schemes](#componentsSecuritySchemes) under the [Components Object](#componentsObject).

Security Requirement Objects that contain multiple schemes require that
all schemes MUST be satisfied for a request to be authorized.
This enables support for scenarios where multiple query parameters or HTTP headers
are required to convey security information.

When a list of Security Requirement Objects is defined on the
[OpenAPI Object](#oasObject) or [Operation Object](#operationObject),
only one of the Security Requirement Objects in the list needs to be satisfied to authorize the request.
"""

"""Patterned Fields"""

# {name}: List[str]
"""
Each name MUST correspond to a security scheme which is declared
in the [Security Schemes](#componentsSecuritySchemes) under the [Components Object](#componentsObject).
If the security scheme is of type `"oauth2"` or `"openIdConnect"`,
then the value is a list of scope names required for the execution,
and the list MAY be empty if authorization does not require a specified scope.
For other security scheme types, the array MAY contain a list of role names which are required for the execution,
but are not otherwise defined or exchanged in-band.
"""



from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Extra, Field
from .discriminator import Discriminator
from .external_documentation import ExternalDocumentation
from .reference import Reference
from .xml import XML


class Schema(BaseModel):
    """
    The Schema Object allows the definition of input and output data types.
    These types can be objects, but also primitives and arrays.
    This object is a superset of
    the [JSON Schema Specification Draft 2020-12](https://tools.ietf.org/html/draft-bhutton-json-schema-00).

    For more information about the properties,
    see [JSON Schema Core](https://tools.ietf.org/html/draft-wright-json-schema-00)
    and [JSON Schema Validation](https://tools.ietf.org/html/draft-wright-json-schema-validation-00).

    Unless stated otherwise, the property definitions follow those of JSON Schema
    and do not add any additional semantics.
    Where JSON Schema indicates that behavior is defined by the application (e.g. for annotations),
    OAS also defers the definition of semantics to the application consuming the OpenAPI document.
    """

    """
    The following properties are taken directly from the 
    [JSON Schema Core](https://tools.ietf.org/html/draft-wright-json-schema-00)
    and follow the same specifications:
    """

    allOf: Optional[List[Union[Reference, "Schema"]]] = None
    """
    This keyword's value MUST be a non-empty array.  Each item of the
    array MUST be a valid JSON Schema.

    An instance validates successfully against this keyword if it
    validates successfully against all schemas defined by this keyword's
    value.
    """

    anyOf: Optional[List[Union[Reference, "Schema"]]] = None
    """
    This keyword's value MUST be a non-empty array.  Each item of the
    array MUST be a valid JSON Schema.

    An instance validates successfully against this keyword if it
    validates successfully against at least one schema defined by this
    keyword's value.  Note that when annotations are being collected, all
    subschemas MUST be examined so that annotations are collected from
    each subschema that validates successfully.
    """

    oneOf: Optional[List[Union[Reference, "Schema"]]] = None
    """
    This keyword's value MUST be a non-empty array.  Each item of the
    array MUST be a valid JSON Schema.

    An instance validates successfully against this keyword if it
    validates successfully against exactly one schema defined by this
    keyword's value.
    """

    schema_not: Optional[Union[Reference, "Schema"]] = Field(default=None, alias="not")
    """
    This keyword's value MUST be a valid JSON Schema.

    An instance is valid against this keyword if it fails to validate
    successfully against the schema defined by this keyword.
    """

    schema_if: Optional[Union[Reference, "Schema"]] = Field(default=None, alias="if")
    """
    This keyword's value MUST be a valid JSON Schema.

    This validation outcome of this keyword's subschema has no direct
    effect on the overall validation result.  Rather, it controls which
    of the "then" or "else" keywords are evaluated.

    Instances that successfully validate against this keyword's subschema
    MUST also be valid against the subschema value of the "then" keyword,
    if present.

    Instances that fail to validate against this keyword's subschema MUST
    also be valid against the subschema value of the "else" keyword, if
    present.

    If annotations (Section 7.7) are being collected, they are collected
    from this keyword's subschema in the usual way, including when the
    keyword is present without either "then" or "else".
    """

    then: Optional[Union[Reference, "Schema"]] = None
    """
    This keyword's value MUST be a valid JSON Schema.

    When "if" is present, and the instance successfully validates against
    its subschema, then validation succeeds against this keyword if the
    instance also successfully validates against this keyword's
    subschema.
   
    This keyword has no effect when "if" is absent, or when the instance
    fails to validate against its subschema.  Implementations MUST NOT
    evaluate the instance against this keyword, for either validation or
    annotation collection purposes, in such cases.
    """

    schema_else: Optional[Union[Reference, "Schema"]] = Field(default=None, alias="else")
    """
    This keyword's value MUST be a valid JSON Schema.

    When "if" is present, and the instance fails to validate against its
    subschema, then validation succeeds against this keyword if the
    instance successfully validates against this keyword's subschema.

    This keyword has no effect when "if" is absent, or when the instance
    successfully validates against its subschema.  Implementations MUST
    NOT evaluate the instance against this keyword, for either validation
    or annotation collection purposes, in such cases.
    """

    dependentSchemas: Optional[Dict[str, Union[Reference, "Schema"]]] = None
    """
    This keyword specifies subschemas that are evaluated if the instance
    is an object and contains a certain property.

    This keyword's value MUST be an object.  Each value in the object
    MUST be a valid JSON Schema.

    If the object key is a property in the instance, the entire instance
    must validate against the subschema.  Its use is dependent on the
    presence of the property.

    Omitting this keyword has the same behavior as an empty object.
    """

    prefixItems: Optional[List[Union[Reference, "Schema"]]] = None
    """
    The value of "prefixItems" MUST be a non-empty array of valid JSON
    Schemas.

    Validation succeeds if each element of the instance validates against
    the schema at the same position, if any.  This keyword does not
    constrain the length of the array.  If the array is longer than this
    keyword's value, this keyword validates only the prefix of matching
    length.

    This keyword produces an annotation value which is the largest index
    to which this keyword applied a subschema.  The value MAY be a
    boolean true if a subschema was applied to every index of the
    instance, such as is produced by the "items" keyword.  This
    annotation affects the behavior of "items" and "unevaluatedItems".

    Omitting this keyword has the same assertion behavior as an empty
    array.
    """

    items: Optional[Union[Reference, "Schema"]] = None
    """
    The value of "items" MUST be a valid JSON Schema.

    This keyword applies its subschema to all instance elements at
    indexes greater than the length of the "prefixItems" array in the
    same schema object, as reported by the annotation result of that
    "prefixItems" keyword.  If no such annotation result exists, "items"
    applies its subschema to all instance array elements.  [[CREF11: Note
    that the behavior of "items" without "prefixItems" is identical to
    that of the schema form of "items" in prior drafts.  When
    "prefixItems" is present, the behavior of "items" is identical to the
    former "additionalItems" keyword.  ]]

    If the "items" subschema is applied to any positions within the
    instance array, it produces an annotation result of boolean true,
    indicating that all remaining array elements have been evaluated
    against this keyword's subschema.

    Omitting this keyword has the same assertion behavior as an empty
    schema.

    Implementations MAY choose to implement or optimize this keyword in
    another way that produces the same effect, such as by directly
    checking for the presence and size of a "prefixItems" array.
    Implementations that do not support annotation collection MUST do so.
    """

    contains: Optional[Union[Reference, "Schema"]] = None
    """
    The value of this keyword MUST be a valid JSON Schema.

    An array instance is valid against "contains" if at least one of its
    elements is valid against the given schema.  The subschema MUST be
    applied to every array element even after the first match has been
    found, in order to collect annotations for use by other keywords.
    This is to ensure that all possible annotations are collected.
    
    Logically, the validation result of applying the value subschema to
    each item in the array MUST be ORed with "false", resulting in an
    overall validation result.

    This keyword produces an annotation value which is an array of the
    indexes to which this keyword validates successfully when applying
    its subschema, in ascending order.  The value MAY be a boolean "true"
    if the subschema validates successfully when applied to every index
    of the instance.  The annotation MUST be present if the instance
    array to which this keyword's schema applies is empty.
    """

    properties: Optional[Dict[str, Union[Reference, "Schema"]]] = None
    """
    The value of "properties" MUST be an object.  Each value of this
    object MUST be a valid JSON Schema.

    Validation succeeds if, for each name that appears in both the
    instance and as a name within this keyword's value, the child
    instance for that name successfully validates against the
    corresponding schema.

    The annotation result of this keyword is the set of instance property
    names matched by this keyword.

    Omitting this keyword has the same assertion behavior as an empty
    object.
    """

    patternProperties: Optional[Dict[str, Union[Reference, "Schema"]]] = None
    """
    The value of "patternProperties" MUST be an object.  Each property
    name of this object SHOULD be a valid regular expression, according
    to the ECMA-262 regular expression dialect.  Each property value of
    this object MUST be a valid JSON Schema.

    Validation succeeds if, for each instance name that matches any
    regular expressions that appear as a property name in this keyword's
    value, the child instance for that name successfully validates
    against each schema that corresponds to a matching regular
    expression.

    The annotation result of this keyword is the set of instance property
    names matched by this keyword.

    Omitting this keyword has the same assertion behavior as an empty
    object.
    """

    additionalProperties: Optional[Union[Reference, "Schema", bool]] = None
    """
    The value of "additionalProperties" MUST be a valid JSON Schema.

    The behavior of this keyword depends on the presence and annotation
    results of "properties" and "patternProperties" within the same
    schema object.  Validation with "additionalProperties" applies only
    to the child values of instance names that do not appear in the
    annotation results of either "properties" or "patternProperties".

    For all such properties, validation succeeds if the child instance
    validates against the "additionalProperties" schema.

    The annotation result of this keyword is the set of instance property
    names validated by this keyword's subschema.

    Omitting this keyword has the same assertion behavior as an empty
    schema.

    Implementations MAY choose to implement or optimize this keyword in
    another way that produces the same effect, such as by directly
    checking the names in "properties" and the patterns in
    "patternProperties" against the instance property set.
    Implementations that do not support annotation collection MUST do so.
    """

    propertyNames: Optional[Union[Reference, "Schema"]] = None
    """
    The value of "propertyNames" MUST be a valid JSON Schema.

    If the instance is an object, this keyword validates if every
    property name in the instance validates against the provided schema.
    Note the property name that the schema is testing will always be a
    string.

    Omitting this keyword has the same behavior as an empty schema.
    """

    unevaluatedItems: Optional[Union[Reference, "Schema"]] = None
    """
    The value of "unevaluatedItems" MUST be a valid JSON Schema.

    The behavior of this keyword depends on the annotation results of
    adjacent keywords that apply to the instance location being
    validated.  Specifically, the annotations from "prefixItems",
    "items", and "contains", which can come from those keywords when they
    are adjacent to the "unevaluatedItems" keyword.  Those three
    annotations, as well as "unevaluatedItems", can also result from any
    and all adjacent in-place applicator (Section 10.2) keywords.  This
    includes but is not limited to the in-place applicators defined in
    this document.

    If no relevant annotations are present, the "unevaluatedItems"
    subschema MUST be applied to all locations in the array.  If a
    boolean true value is present from any of the relevant annotations,
    "unevaluatedItems" MUST be ignored.  Otherwise, the subschema MUST be
    applied to any index greater than the largest annotation value for
    "prefixItems", which does not appear in any annotation value for
    "contains".

    This means that "prefixItems", "items", "contains", and all in-place
    applicators MUST be evaluated before this keyword can be evaluated.
    Authors of extension keywords MUST NOT define an in-place applicator
    that would need to be evaluated after this keyword.

    If the "unevaluatedItems" subschema is applied to any positions
    within the instance array, it produces an annotation result of
    boolean true, analogous to the behavior of "items".

    Omitting this keyword has the same assertion behavior as an empty
    schema.
    """

    unevaluatedProperties: Optional[Union[Reference, "Schema"]] = None
    """
    The value of "unevaluatedProperties" MUST be a valid JSON Schema.

    The behavior of this keyword depends on the annotation results of
    adjacent keywords that apply to the instance location being
    validated.  Specifically, the annotations from "properties",
    "patternProperties", and "additionalProperties", which can come from
    those keywords when they are adjacent to the "unevaluatedProperties"
    keyword.  Those three annotations, as well as
    "unevaluatedProperties", can also result from any and all adjacent
    in-place applicator (Section 10.2) keywords.  This includes but is
    not limited to the in-place applicators defined in this document.

    Validation with "unevaluatedProperties" applies only to the child
    values of instance names that do not appear in the "properties",
    "patternProperties", "additionalProperties", or
    "unevaluatedProperties" annotation results that apply to the instance
    location being validated.

    For all such properties, validation succeeds if the child instance
    validates against the "unevaluatedProperties" schema.

    This means that "properties", "patternProperties",
    "additionalProperties", and all in-place applicators MUST be
    evaluated before this keyword can be evaluated.  Authors of extension
    keywords MUST NOT define an in-place applicator that would need to be
    evaluated after this keyword.

    The annotation result of this keyword is the set of instance property
    names validated by this keyword's subschema.

    Omitting this keyword has the same assertion behavior as an empty
    schema.
    """

    """
    The following properties are taken directly from the 
    [JSON Schema Validation](https://tools.ietf.org/html/draft-wright-json-schema-validation-00)
    and follow the same specifications:
    """

    type: Optional[Union[str, List[str]]] = None
    """
    The value of this keyword MUST be either a string or an array.  If it
    is an array, elements of the array MUST be strings and MUST be
    unique.

    String values MUST be one of the six primitive types ("null",
    "boolean", "object", "array", "number", or "string"), or "integer"
    which matches any number with a zero fractional part.
    
    An instance validates if and only if the instance is in any of the
    sets listed for this keyword.
    """

    enum: Optional[List[Any]] = Field(default=None, min_items=1)
    """
    The value of this keyword MUST be an array.  This array SHOULD have
    at least one element.  Elements in the array SHOULD be unique.
    
    An instance validates successfully against this keyword if its value
    is equal to one of the elements in this keyword's array value.
    
    Elements in the array might be of any type, including null.
    """

    const: Optional[Any] = None
    """
    The value of this keyword MAY be of any type, including null.
    
    Use of this keyword is functionally equivalent to an "enum"
    (Section 6.1.2) with a single value.

    An instance validates successfully against this keyword if its value
    is equal to the value of the keyword.
    """

    multipleOf: Optional[float] = Field(default=None, gt=0.0)
    """
    The value of "multipleOf" MUST be a number, strictly greater than 0.

    A numeric instance is only valid if division by this keyword's value
    results in an integer.
    """

    maximum: Optional[float] = None
    """
    The value of "maximum" MUST be a number, representing an inclusive
    upper limit for a numeric instance.

    If the instance is a number, then this keyword validates only if the
    instance is less than or exactly equal to "maximum".
    """

    exclusiveMaximum: Optional[float] = None
    """
    The value of "exclusiveMaximum" MUST be a number, representing an
    exclusive upper limit for a numeric instance.

    If the instance is a number, then the instance is valid only if it
    has a value strictly less than (not equal to) "exclusiveMaximum".
    """

    minimum: Optional[float] = None
    """
    The value of "minimum" MUST be a number, representing an inclusive
    lower limit for a numeric instance.

    If the instance is a number, then this keyword validates only if the
    instance is greater than or exactly equal to "minimum".
    """

    exclusiveMinimum: Optional[float] = None
    """
    The value of "exclusiveMinimum" MUST be a number, representing an
    exclusive lower limit for a numeric instance.

    If the instance is a number, then the instance is valid only if it
    has a value strictly greater than (not equal to) "exclusiveMinimum".
    """

    maxLength: Optional[int] = Field(default=None, ge=0)
    """
    The value of this keyword MUST be a non-negative integer.

    A string instance is valid against this keyword if its length is less
    than, or equal to, the value of this keyword.

    The length of a string instance is defined as the number of its
    characters as defined by RFC 8259 [RFC8259].
    """

    minLength: Optional[int] = Field(default=None, ge=0)
    """
    The value of this keyword MUST be a non-negative integer.

    A string instance is valid against this keyword if its length is
    greater than, or equal to, the value of this keyword.

    The length of a string instance is defined as the number of its
    characters as defined by RFC 8259 [RFC8259].

    Omitting this keyword has the same behavior as a value of 0.
    """

    pattern: Optional[str] = None
    """
    The value of this keyword MUST be a string.  This string SHOULD be a
    valid regular expression, according to the ECMA-262 regular
    expression dialect.

    A string instance is considered valid if the regular expression
    matches the instance successfully.  Recall: regular expressions are
    not implicitly anchored.
    """

    maxItems: Optional[int] = Field(default=None, ge=0)
    """
    The value of this keyword MUST be a non-negative integer.

    An array instance is valid against "maxItems" if its size is less
    than, or equal to, the value of this keyword.
    """

    minItems: Optional[int] = Field(default=None, ge=0)
    """
    The value of this keyword MUST be a non-negative integer.

    An array instance is valid against "minItems" if its size is greater
    than, or equal to, the value of this keyword.

    Omitting this keyword has the same behavior as a value of 0.
    """

    uniqueItems: Optional[bool] = None
    """
    The value of this keyword MUST be a boolean.

    If this keyword has boolean value false, the instance validates
    successfully.  If it has boolean value true, the instance validates
    successfully if all of its elements are unique.

    Omitting this keyword has the same behavior as a value of false.
    """

    maxContains: Optional[int] = Field(default=None, ge=0)
    """
    The value of this keyword MUST be a non-negative integer.

    If "contains" is not present within the same schema object, then this
    keyword has no effect.

    An instance array is valid against "maxContains" in two ways,
    depending on the form of the annotation result of an adjacent
    "contains" [json-schema] keyword.  The first way is if the annotation
    result is an array and the length of that array is less than or equal
    to the "maxContains" value.  The second way is if the annotation
    result is a boolean "true" and the instance array length is less than
    or equal to the "maxContains" value.
    """

    minContains: Optional[int] = Field(default=None, ge=0)
    """
    The value of this keyword MUST be a non-negative integer.

    If "contains" is not present within the same schema object, then this
    keyword has no effect.

    An instance array is valid against "minContains" in two ways,
    depending on the form of the annotation result of an adjacent
    "contains" [json-schema] keyword.  The first way is if the annotation
    result is an array and the length of that array is greater than or
    equal to the "minContains" value.  The second way is if the
    annotation result is a boolean "true" and the instance array length
    is greater than or equal to the "minContains" value.

    A value of 0 is allowed, but is only useful for setting a range of
    occurrences from 0 to the value of "maxContains".  A value of 0 with
    no "maxContains" causes "contains" to always pass validation.

    Omitting this keyword has the same behavior as a value of 1.
    """

    maxProperties: Optional[int] = Field(default=None, ge=0)
    """
    The value of this keyword MUST be a non-negative integer.

    An object instance is valid against "maxProperties" if its number of
    properties is less than, or equal to, the value of this keyword.
    """

    minProperties: Optional[int] = Field(default=None, ge=0)
    """
    The value of this keyword MUST be a non-negative integer.

    An object instance is valid against "minProperties" if its number of
    properties is greater than, or equal to, the value of this keyword.

    Omitting this keyword has the same behavior as a value of 0.
    """

    required: Optional[List[str]] = None
    """
    The value of this keyword MUST be an array.  Elements of this array,
    if any, MUST be strings, and MUST be unique.

    An object instance is valid against this keyword if every item in the
    array is the name of a property in the instance.

    Omitting this keyword has the same behavior as an empty array.
    """

    dependentRequired: Optional[Dict[str, List[str]]] = None
    """
    The value of this keyword MUST be an object.  Properties in this
    object, if any, MUST be arrays.  Elements in each array, if any, MUST
    be strings, and MUST be unique.

    This keyword specifies properties that are required if a specific
    other property is present.  Their requirement is dependent on the
    presence of the other property.

    Validation succeeds if, for each name that appears in both the
    instance and as a name within this keyword's value, every item in the
    corresponding array is also the name of a property in the instance.

    Omitting this keyword has the same behavior as an empty object.
    """

    schema_format: Optional[str] = Field(default=None, alias="format")
    """
    From OpenAPI:
    See [Data Type Formats](#dataTypeFormat) for further details.
    While relying on JSON Schema's defined formats, the OAS offers a few additional predefined formats.
    
    From JSON Schema:
    Structural validation alone may be insufficient to allow an
    application to correctly utilize certain values.  The "format"
    annotation keyword is defined to allow schema authors to convey
    semantic information for a fixed subset of values which are
    accurately described by authoritative resources, be they RFCs or
    other external specifications.

    The value of this keyword is called a format attribute.  It MUST be a
    string.  A format attribute can generally only validate a given set
    of instance types.  If the type of the instance to validate is not in
    this set, validation for this format attribute and instance SHOULD
    succeed.  All format attributes defined in this section apply to
    strings, but a format attribute can be specified to apply to any
    instance types defined in the data model defined in the core JSON
    Schema. [json-schema] [[CREF1: Note that the "type" keyword in this
    specification defines an "integer" type which is not part of the data
    model.  Therefore a format attribute can be limited to numbers, but
    not specifically to integers.  However, a numeric format can be used
    alongside the "type" keyword with a value of "integer", or could be
    explicitly defined to always pass if the number is not an integer,
    which produces essentially the same behavior as only applying to
    integers.  ]]
    """

    contentEncoding: Optional[str] = None
    """
    If the instance value is a string, this property defines that the
    string SHOULD be interpreted as binary data and decoded using the
    encoding named by this property.

    Possible values indicating base 16, 32, and 64 encodings with several
    variations are listed in RFC 4648 [RFC4648].  Additionally, sections
    6.7 and 6.8 of RFC 2045 [RFC2045] provide encodings used in MIME.  As
    "base64" is defined in both RFCs, the definition from RFC 4648 SHOULD
    be assumed unless the string is specifically intended for use in a
    MIME context.  Note that all of these encodings result in strings
    consisting only of 7-bit ASCII characters.  Therefore, this keyword
    has no meaning for strings containing characters outside of that
    range.

    If this keyword is absent, but "contentMediaType" is present, this
    indicates that the encoding is the identity encoding, meaning that no
    transformation was needed in order to represent the content in a
    UTF-8 string.
    """

    contentMediaType: Optional[str] = None
    """
    If the instance is a string, this property indicates the media type
    of the contents of the string.  If "contentEncoding" is present, this
    property describes the decoded string.

    The value of this property MUST be a string, which MUST be a media
    type, as defined by RFC 2046 [RFC2046].
    """

    contentSchema: Optional[Union[Reference, "Schema"]] = None
    """
    If the instance is a string, and if "contentMediaType" is present,
    this property contains a schema which describes the structure of the
    string.

    This keyword MAY be used with any media type that can be mapped into
    JSON Schema's data model.

    The value of this property MUST be a valid JSON schema.  It SHOULD be
    ignored if "contentMediaType" is not present.
    """

    title: Optional[str] = None
    """
    The value of "title" MUST be a string.

    The title can be used to decorate a user interface with
    information about the data produced by this user interface.
    A title will preferably be short.
    """

    description: Optional[str] = None
    """
    From OpenAPI:
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text representation.
    
    From JSON Schema:
    The value "description" MUST be a string.

    The description can be used to decorate a user interface with
    information about the data produced by this user interface.
    A description will provide explanation about the purpose of
    the instance described by this schema.
    """

    default: Optional[Any] = None
    """
    There are no restrictions placed on the value of this keyword.  When
    multiple occurrences of this keyword are applicable to a single sub-
    instance, implementations SHOULD remove duplicates.

    This keyword can be used to supply a default JSON value associated
    with a particular schema.  It is RECOMMENDED that a default value be
    valid against the associated schema.
    """

    deprecated: Optional[bool] = None
    """ 
    The value of this keyword MUST be a boolean.  When multiple
    occurrences of this keyword are applicable to a single sub-instance,
    applications SHOULD consider the instance location to be deprecated
    if any occurrence specifies a true value.

    If "deprecated" has a value of boolean true, it indicates that
    applications SHOULD refrain from usage of the declared property.  It
    MAY mean the property is going to be removed in the future.

    A root schema containing "deprecated" with a value of true indicates
    that the entire resource being described MAY be removed in the
    future.

    The "deprecated" keyword applies to each instance location to which
    the schema object containing the keyword successfully applies.  This
    can result in scenarios where every array item or object property is
    deprecated even though the containing array or object is not.

    Omitting this keyword has the same behavior as a value of false.
    """

    readOnly: Optional[bool] = None
    """
    The value of "readOnly" MUST be a boolean.  When multiple
    occurrences of this keyword are applicable to a single sub-instance,
    the resulting behavior SHOULD be as for a true value if any
    occurrence specifies a true value, and SHOULD be as for a false value
    otherwise.

    If "readOnly" has a value of boolean true, it indicates that the
    value of the instance is managed exclusively by the owning authority,
    and attempts by an application to modify the value of this property
    are expected to be ignored or rejected by that owning authority.

    An instance document that is marked as "readOnly" for the entire
    document MAY be ignored if sent to the owning authority, or MAY
    result in an error, at the authority's discretion.

    For example, "readOnly" would be used to mark a database-generated
    serial number as read-only, while "writeOnly" would be used to mark a
    password input field.

    This keyword can be used to assist in user interface instance
    generation.  In particular, an application MAY choose to use a widget
    that hides input values as they are typed for write-only fields.

    Omitting these keywords has the same behavior as values of false.
    """

    writeOnly: Optional[bool] = None
    """
    The value of "writeOnly" MUST be a boolean.  When multiple
    occurrences of this keyword are applicable to a single sub-instance,
    the resulting behavior SHOULD be as for a true value if any
    occurrence specifies a true value, and SHOULD be as for a false value
    otherwise.

    If "writeOnly" has a value of boolean true, it indicates that the
    value is never present when the instance is retrieved from the owning
    authority.  It can be present when sent to the owning authority to
    update or create the document (or the resource it represents), but it
    will not be included in any updated or newly created version of the
    instance.

    An instance document that is marked as "writeOnly" for the entire
    document MAY be returned as a blank document of some sort, or MAY
    produce an error upon retrieval, or have the retrieval request
    ignored, at the authority's discretion.

    For example, "readOnly" would be used to mark a database-generated
    serial number as read-only, while "writeOnly" would be used to mark a
    password input field.

    This keyword can be used to assist in user interface instance
    generation.  In particular, an application MAY choose to use a widget
    that hides input values as they are typed for write-only fields.

    Omitting these keywords has the same behavior as values of false.
    """

    examples: Optional[List[Any]] = None
    """
    The value of this keyword MUST be an array.  There are no
    restrictions placed on the values within the array.  When multiple
    occurrences of this keyword are applicable to a single sub-instance,
    implementations MUST provide a flat array of all values rather than
    an array of arrays.
    
    This keyword can be used to provide sample JSON values associated
    with a particular schema, for the purpose of illustrating usage.  It
    is RECOMMENDED that these values be valid against the associated
    schema.

    Implementations MAY use the value(s) of "default", if present, as an
    additional example.  If "examples" is absent, "default" MAY still be
    used in this manner.
    """

    """
    The OpenAPI Specification's base vocabulary is comprised of the following keywords:
    """

    discriminator: Optional[Discriminator] = None
    """
    Adds support for polymorphism.
    The discriminator is an object name that is used to differentiate between other schemas
    which may satisfy the payload description.
    See [Composition and Inheritance](#schemaComposition) for more details.
    """

    xml: Optional[XML] = None
    """
    This MAY be used only on properties schemas.
    It has no effect on root schemas.
    Adds additional metadata to describe the XML representation of this property.
    """

    externalDocs: Optional[ExternalDocumentation] = None
    """
    Additional external documentation for this schema.
    """

    example: Optional[Any] = None
    """
    A free-form property to include an example of an instance for this schema.
    To represent examples that cannot be naturally represented in JSON or YAML,
    a string value can be used to contain the example with escaping where necessary.
    
    Deprecated: The example property has been deprecated in favor of the JSON Schema examples keyword.
    Use of example is discouraged, and later versions of this specification may remove it.
    """

    class Config:
        extra = Extra.ignore
        allow_population_by_field_name = True
        schema_extra = {
            "examples": [
                {"type": "string", "format": "email"},
                {
                    "type": "object",
                    "required": ["name"],
                    "properties": {
                        "name": {"type": "string"},
                        "address": {"$ref": "#/components/schemas/Address"},
                        "age": {"type": "integer", "format": "int32", "minimum": 0},
                    },
                },
                {"type": "object", "additionalProperties": {"type": "string"}},
                {"type": "object", "additionalProperties": {"$ref": "#/components/schemas/ComplexModel"}},
                {
                    "type": "object",
                    "properties": {"id": {"type": "integer", "format": "int64"}, "name": {"type": "string"}},
                    "required": ["name"],
                    "example": {"name": "Puma", "id": 1},
                },
                {
                    "type": "object",
                    "required": ["message", "code"],
                    "properties": {
                        "message": {"type": "string"},
                        "code": {"type": "integer", "minimum": 100, "maximum": 600},
                    },
                },
                {
                    "allOf": [
                        {"$ref": "#/components/schemas/ErrorModel"},
                        {"type": "object", "required": ["rootCause"], "properties": {"rootCause": {"type": "string"}}},
                    ]
                },
                {
                    "type": "object",
                    "discriminator": {"propertyName": "petType"},
                    "properties": {"name": {"type": "string"}, "petType": {"type": "string"}},
                    "required": ["name", "petType"],
                },
                {
                    "description": "A representation of a cat. "
                    "Note that `Cat` will be used as the discriminator value.",
                    "allOf": [
                        {"$ref": "#/components/schemas/Pet"},
                        {
                            "type": "object",
                            "properties": {
                                "huntingSkill": {
                                    "type": "string",
                                    "description": "The measured skill for hunting",
                                    "default": "lazy",
                                    "enum": ["clueless", "lazy", "adventurous", "aggressive"],
                                }
                            },
                            "required": ["huntingSkill"],
                        },
                    ],
                },
                {
                    "description": "A representation of a dog. "
                    "Note that `Dog` will be used as the discriminator value.",
                    "allOf": [
                        {"$ref": "#/components/schemas/Pet"},
                        {
                            "type": "object",
                            "properties": {
                                "packSize": {
                                    "type": "integer",
                                    "format": "int32",
                                    "description": "the size of the pack the dog is from",
                                    "default": 0,
                                    "minimum": 0,
                                }
                            },
                            "required": ["packSize"],
                        },
                    ],
                },
            ]
        }



from typing import Dict, Optional

from pydantic import BaseModel, Extra


class Discriminator(BaseModel):
    """
    When request bodies or response payloads may be one of a number of different schemas,
    a `discriminator` object can be used to aid in serialization, deserialization, and validation.

    The discriminator is a specific object in a schema which is used to inform the consumer of the specification
    of an alternative schema based on the value associated with it.

    When using the discriminator, _inline_ schemas will not be considered.
    """

    propertyName: str = ...
    """
    **REQUIRED**. The name of the property in the payload that will hold the discriminator value.
    """

    mapping: Optional[Dict[str, str]] = None
    """
    An object to hold mappings between payload values and schema names or references.
    """

    class Config:
        extra = Extra.ignore
        schema_extra = {
            "examples": [
                {
                    "propertyName": "petType",
                    "mapping": {
                        "dog": "#/components/schemas/Dog",
                        "monster": "https://gigantic-server.com/schemas/Monster/schema.json",
                    },
                }
            ]
        }



from typing import Optional

from pydantic import BaseModel, Extra

from .external_documentation import ExternalDocumentation


class Tag(BaseModel):
    """
    Adds metadata to a single tag that is used by the [Operation Object](#operationObject).
    It is not mandatory to have a Tag Object per tag defined in the Operation Object instances.
    """

    name: str = ...
    """
    **REQUIRED**. The name of the tag.
    """

    description: Optional[str] = None
    """
    A short description for the tag.
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text representation.
    """

    externalDocs: Optional[ExternalDocumentation] = None
    """
    Additional external documentation for this tag.
    """

    class Config:
        extra = Extra.ignore
        schema_extra = {"examples": [{"name": "pet", "description": "Pets operations"}]}



from typing import Optional

from pydantic import BaseModel, Extra


class XML(BaseModel):
    """
    A metadata object that allows for more fine-tuned XML model definitions.

    When using arrays, XML element names are *not* inferred (for singular/plural forms)
    and the `name` property SHOULD be used to add that information.
    See examples for expected behavior.
    """

    name: Optional[str] = None
    """
    Replaces the name of the element/attribute used for the described schema property.
    When defined within `items`, it will affect the name of the individual XML elements within the list.
    When defined alongside `type` being `array` (outside the `items`),
    it will affect the wrapping element and only if `wrapped` is `true`.
    If `wrapped` is `false`, it will be ignored.
    """

    namespace: Optional[str] = None
    """
    The URI of the namespace definition.
    Value MUST be in the form of an absolute URI.
    """

    prefix: Optional[str] = None
    """
    The prefix to be used for the [name](#xmlName).
    """

    attribute: bool = False
    """
    Declares whether the property definition translates to an attribute instead of an element.
    Default value is `false`.
    """

    wrapped: bool = False
    """
    MAY be used only for an array definition.
    Signifies whether the array is wrapped (for example, `<books><book/><book/></books>`)
    or unwrapped (`<book/><book/>`).
    Default value is `false`.
    The definition takes effect only when defined alongside `type` being `array` (outside the `items`).
    """

    class Config:
        extra = Extra.ignore
        schema_extra = {
            "examples": [
                {"namespace": "http://example.com/schema/sample", "prefix": "sample"},
                {"name": "aliens", "wrapped": True},
            ]
        }



from typing import Optional

from pydantic import BaseModel, Extra, Field


class Reference(BaseModel):
    """
    A simple object to allow referencing other components in the OpenAPI document, internally and externally.

    The `$ref` string value contains a URI [RFC3986](https://tools.ietf.org/html/rfc3986),
    which identifies the location of the value being referenced.

    See the rules for resolving [Relative References](#relativeReferencesURI).
    """

    ref: str = Field(alias="$ref")
    """**REQUIRED**. The reference identifier. This MUST be in the form of a URI."""

    summary: Optional[str] = None
    """
    A short summary which by default SHOULD override that of the referenced component.
    If the referenced object-type does not allow a `summary` field, then this field has no effect.
    """

    description: Optional[str] = None
    """
    A description which by default SHOULD override that of the referenced component.
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text representation.
    If the referenced object-type does not allow a `description` field, then this field has no effect.
    """

    class Config:
        extra = Extra.ignore
        allow_population_by_field_name = True
        schema_extra = {
            "examples": [{"$ref": "#/components/schemas/Pet"}, {"$ref": "Pet.json"}, {"$ref": "definitions.json#/Pet"}]
        }



from typing import Dict, Optional

from pydantic import BaseModel, Extra

from .media_type import MediaType


class RequestBody(BaseModel):
    """Describes a single request body."""

    description: Optional[str] = None
    """
    A brief description of the request body.
    This could contain examples of use.  
    
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text representation.
    """

    content: Dict[str, MediaType] = ...
    """
    **REQUIRED**. The content of the request body.
    The key is a media type or [media type range](https://tools.ietf.org/html/rfc7231#appendix-D)
    and the value describes it.
    
    For requests that match multiple keys, only the most specific key is applicable. e.g. text/plain overrides text/*
    """

    required: bool = False
    """
    Determines if the request body is required in the request. Defaults to `false`.
    """

    class Config:
        extra = Extra.ignore
        schema_extra = {
            "examples": [
                {
                    "description": "user to add to the system",
                    "content": {
                        "application/json": {
                            "schema": {"$ref": "#/components/schemas/User"},
                            "examples": {
                                "user": {
                                    "summary": "User Example",
                                    "externalValue": "http://foo.bar/examples/user-example.json",
                                }
                            },
                        },
                        "application/xml": {
                            "schema": {"$ref": "#/components/schemas/User"},
                            "examples": {
                                "user": {
                                    "summary": "User example in XML",
                                    "externalValue": "http://foo.bar/examples/user-example.xml",
                                }
                            },
                        },
                        "text/plain": {
                            "examples": {
                                "user": {
                                    "summary": "User example in Plain text",
                                    "externalValue": "http://foo.bar/examples/user-example.txt",
                                }
                            }
                        },
                        "*/*": {
                            "examples": {
                                "user": {
                                    "summary": "User example in other format",
                                    "externalValue": "http://foo.bar/examples/user-example.whatever",
                                }
                            }
                        },
                    },
                },
                {
                    "description": "user to add to the system",
                    "content": {"text/plain": {"schema": {"type": "array", "items": {"type": "string"}}}},
                },
            ]
        }



from typing import Any, Dict, Optional, Union

from pydantic import BaseModel, Field, Extra

from .example import Example
from .media_type import MediaType
from .reference import Reference
from .schema import Schema


class Parameter(BaseModel):
    """
    Describes a single operation parameter.

    A unique parameter is defined by a combination of a [name](#parameterName) and [location](#parameterIn).
    """

    """Fixed Fields"""

    name: str = ...
    """
    **REQUIRED**. The name of the parameter.
    Parameter names are *case sensitive*. 
    
    - If [`in`](#parameterIn) is `"path"`, the `name` field MUST correspond to a template expression occurring 
      within the [path](#pathsPath) field in the [Paths Object](#pathsObject).
      See [Path Templating](#pathTemplating) for further information.
    - If [`in`](#parameterIn) is `"header"` and the `name` field is `"Accept"`, `"Content-Type"` or `"Authorization"`, 
      the parameter definition SHALL be ignored.
    - For all other cases, the `name` corresponds to the parameter name used by the [`in`](#parameterIn) property.
    """

    param_in: str = Field(alias="in")
    """
    **REQUIRED**. The location of the parameter. Possible values are `"query"`, `"header"`, `"path"` or `"cookie"`.
    """

    description: Optional[str] = None
    """
    A brief description of the parameter.
    This could contain examples of use.
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text representation.
    """

    required: bool = False
    """
    Determines whether this parameter is mandatory.
    If the [parameter location](#parameterIn) is `"path"`, this property is **REQUIRED** and its value MUST be `true`.
    Otherwise, the property MAY be included and its default value is `false`.
    """

    deprecated: bool = False
    """
    Specifies that a parameter is deprecated and SHOULD be transitioned out of usage.
    Default value is `false`.
    """

    allowEmptyValue: bool = False
    """
    Sets the ability to pass empty-valued parameters.
    This is valid only for `query` parameters and allows sending a parameter with an empty value.
    Default value is `false`.
    If [`style`](#parameterStyle) is used, and if behavior is `n/a` (cannot be serialized),
    the value of `allowEmptyValue` SHALL be ignored.
    Use of this property is NOT RECOMMENDED, as it is likely to be removed in a later revision.
    """

    """
    The rules for serialization of the parameter are specified in one of two ways.
    For simpler scenarios, a [`schema`](#parameterSchema) and [`style`](#parameterStyle)
    can describe the structure and syntax of the parameter.
    """

    style: Optional[str] = None
    """
    Describes how the parameter value will be serialized depending on the type of the parameter value.
    Default values (based on value of `in`):
    
    - for `query` - `form`;
    - for `path` - `simple`;
    - for `header` - `simple`;
    - for `cookie` - `form`.
    """

    explode: bool = False
    """
    When this is true, parameter values of type `array` or `object` generate separate parameters
    for each value of the array or key-value pair of the map.
    For other types of parameters this property has no effect.
    When [`style`](#parameterStyle) is `form`, the default value is `true`.
    For all other styles, the default value is `false`.
    """

    allowReserved: bool = False
    """
    Determines whether the parameter value SHOULD allow reserved characters,
    as defined by [RFC3986](https://tools.ietf.org/html/rfc3986#section-2.2)
    `:/?#[]@!$&'()*+,;=` to be included without percent-encoding.
    This property only applies to parameters with an `in` value of `query`.
    The default value is `false`.
    """

    param_schema: Optional[Union[Reference, Schema]] = Field(default=None, alias="schema")
    """
    The schema defining the type used for the parameter.
    """

    example: Optional[Any] = None
    """
    Example of the parameter's potential value.
    The example SHOULD match the specified schema and encoding properties if present.
    The `example` field is mutually exclusive of the `examples` field.
    Furthermore, if referencing a `schema` that contains an example, 
    the `example` value SHALL _override_ the example provided by the schema.
    To represent examples of media types that cannot naturally be represented in JSON or YAML,
    a string value can contain the example with escaping where necessary.
    """

    examples: Optional[Dict[str, Union[Example, Reference]]] = None
    """
    Examples of the parameter's potential value.
    Each example SHOULD contain a value in the correct format as specified in the parameter encoding.
    The `examples` field is mutually exclusive of the `example` field.
    Furthermore, if referencing a `schema` that contains an example,
    the `examples` value SHALL _override_ the example provided by the schema.
    """

    """
    For more complex scenarios, the [`content`](#parameterContent) property 
    can define the media type and schema of the parameter.
    A parameter MUST contain either a `schema` property, or a `content` property, but not both.
    When `example` or `examples` are provided in conjunction with the `schema` object,
    the example MUST follow the prescribed serialization strategy for the parameter.
    """

    content: Optional[Dict[str, MediaType]] = None
    """
    A map containing the representations for the parameter.
    The key is the media type and the value describes it.
    The map MUST only contain one entry.
    """

    class Config:
        extra = Extra.ignore
        allow_population_by_field_name = True
        schema_extra = {
            "examples": [
                {
                    "name": "token",
                    "in": "header",
                    "description": "token to be passed as a header",
                    "required": True,
                    "schema": {"type": "array", "items": {"type": "integer", "format": "int64"}},
                    "style": "simple",
                },
                {
                    "name": "username",
                    "in": "path",
                    "description": "username to fetch",
                    "required": True,
                    "schema": {"type": "string"},
                },
                {
                    "name": "id",
                    "in": "query",
                    "description": "ID of the object to fetch",
                    "required": False,
                    "schema": {"type": "array", "items": {"type": "string"}},
                    "style": "form",
                    "explode": True,
                },
                {
                    "in": "query",
                    "name": "freeForm",
                    "schema": {"type": "object", "additionalProperties": {"type": "integer"}},
                    "style": "form",
                },
                {
                    "in": "query",
                    "name": "coordinates",
                    "content": {
                        "application/json": {
                            "schema": {
                                "type": "object",
                                "required": ["lat", "long"],
                                "properties": {"lat": {"type": "number"}, "long": {"type": "number"}},
                            }
                        }
                    },
                },
            ]
        }



"""
OpenAPI v3.1.0 schema types, created according to the specification:
https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.1.0.md

The type orders are according to the contents of the specification:
https://github.com/OAI/OpenAPI-Specification/blob/master/versions/3.1.0.md#table-of-contents
"""

from .open_api import OpenAPI
from .info import Info
from .contact import Contact
from .license import License
from .server import Server
from .server_variable import ServerVariable
from .components import Components
from .paths import Paths
from .path_item import PathItem
from .operation import Operation
from .external_documentation import ExternalDocumentation
from .parameter import Parameter
from .request_body import RequestBody
from .media_type import MediaType
from .encoding import Encoding
from .responses import Responses
from .response import Response
from .callback import Callback
from .example import Example
from .link import Link
from .header import Header
from .tag import Tag
from .reference import Reference
from .schema import Schema
from .discriminator import Discriminator
from .xml import XML
from .security_scheme import SecurityScheme
from .oauth_flows import OAuthFlows
from .oauth_flow import OAuthFlow
from .security_requirement import SecurityRequirement

# resolve forward references
Encoding.update_forward_refs(Header=Header)
Schema.update_forward_refs()



from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Extra, Field
from .discriminator import Discriminator
from .external_documentation import ExternalDocumentation
from .reference import Reference
from .xml import XML


class Schema(BaseModel):
    """
    The Schema Object allows the definition of input and output data types.
    These types can be objects, but also primitives and arrays.
    This object is an extended subset of the [JSON Schema Specification Wright Draft 00](https://json-schema.org/).

    For more information about the properties,
    see [JSON Schema Core](https://tools.ietf.org/html/draft-wright-json-schema-00)
    and [JSON Schema Validation](https://tools.ietf.org/html/draft-wright-json-schema-validation-00).
    Unless stated otherwise, the property definitions follow the JSON Schema.
    """

    """
    The following properties are taken directly from the JSON Schema definition and follow the same specifications:
    """

    title: Optional[str] = None
    """
    The value of "title" MUST be a string.

    The title can be used to decorate a user interface with
    information about the data produced by this user interface.
    The title will preferrably be short.
    """

    multipleOf: Optional[float] = Field(default=None, gt=0.0)
    """
    The value of "multipleOf" MUST be a number, strictly greater than 0.
    
    A numeric instance is only valid if division by this keyword's value
    results in an integer.
    """

    maximum: Optional[float] = None
    """
    The value of "maximum" MUST be a number, representing an upper limit
    for a numeric instance.
    
    If the instance is a number, then this keyword validates if
    "exclusiveMaximum" is true and instance is less than the provided
    value, or else if the instance is less than or exactly equal to the
    provided value.
    """

    exclusiveMaximum: Optional[bool] = None
    """
    The value of "exclusiveMaximum" MUST be a boolean, representing
    whether the limit in "maximum" is exclusive or not.  An undefined
    value is the same as false.
    
    If "exclusiveMaximum" is true, then a numeric instance SHOULD NOT be
    equal to the value specified in "maximum".  If "exclusiveMaximum" is
    false (or not specified), then a numeric instance MAY be equal to the
    value of "maximum".
    """

    minimum: Optional[float] = None
    """
    The value of "minimum" MUST be a number, representing a lower limit
    for a numeric instance.
    
    If the instance is a number, then this keyword validates if
    "exclusiveMinimum" is true and instance is greater than the provided
    value, or else if the instance is greater than or exactly equal to
    the provided value.
    """

    exclusiveMinimum: Optional[bool] = None
    """
    The value of "exclusiveMinimum" MUST be a boolean, representing
    whether the limit in "minimum" is exclusive or not.  An undefined
    value is the same as false.
    
    If "exclusiveMinimum" is true, then a numeric instance SHOULD NOT be
    equal to the value specified in "minimum".  If "exclusiveMinimum" is
    false (or not specified), then a numeric instance MAY be equal to the
    value of "minimum".
    """

    maxLength: Optional[int] = Field(default=None, ge=0)
    """
    The value of this keyword MUST be a non-negative integer.

    The value of this keyword MUST be an integer.  This integer MUST be
    greater than, or equal to, 0.
    
    A string instance is valid against this keyword if its length is less
    than, or equal to, the value of this keyword.
    
    The length of a string instance is defined as the number of its
    characters as defined by RFC 7159 [RFC7159].
    """

    minLength: Optional[int] = Field(default=None, ge=0)
    """
    A string instance is valid against this keyword if its length is
    greater than, or equal to, the value of this keyword.
    
    The length of a string instance is defined as the number of its
    characters as defined by RFC 7159 [RFC7159].
    
    The value of this keyword MUST be an integer.  This integer MUST be
    greater than, or equal to, 0.
    
    "minLength", if absent, may be considered as being present with
    integer value 0.
    """

    pattern: Optional[str] = None
    """
    The value of this keyword MUST be a string.  This string SHOULD be a
    valid regular expression, according to the ECMA 262 regular
    expression dialect.
    
    A string instance is considered valid if the regular expression
    matches the instance successfully.  Recall: regular expressions are
    not implicitly anchored.
    """

    maxItems: Optional[int] = Field(default=None, ge=0)
    """
    The value of this keyword MUST be an integer.  This integer MUST be
    greater than, or equal to, 0.
    
    An array instance is valid against "maxItems" if its size is less
    than, or equal to, the value of this keyword.
    """

    minItems: Optional[int] = Field(default=None, ge=0)
    """
    The value of this keyword MUST be an integer.  This integer MUST be
    greater than, or equal to, 0.
    
    An array instance is valid against "minItems" if its size is greater
    than, or equal to, the value of this keyword.
    
    If this keyword is not present, it may be considered present with a
    value of 0.
    """

    uniqueItems: Optional[bool] = None
    """
    The value of this keyword MUST be a boolean.

    If this keyword has boolean value false, the instance validates
    successfully.  If it has boolean value true, the instance validates
    successfully if all of its elements are unique.
    
    If not present, this keyword may be considered present with boolean
    value false.
    """

    maxProperties: Optional[int] = Field(default=None, ge=0)
    """
    The value of this keyword MUST be an integer.  This integer MUST be
    greater than, or equal to, 0.
    
    An object instance is valid against "maxProperties" if its number of
    properties is less than, or equal to, the value of this keyword.
    """

    minProperties: Optional[int] = Field(default=None, ge=0)
    """
    The value of this keyword MUST be an integer.  This integer MUST be
    greater than, or equal to, 0.
    
    An object instance is valid against "minProperties" if its number of
    properties is greater than, or equal to, the value of this keyword.
    
    If this keyword is not present, it may be considered present with a
    value of 0.
    """

    required: Optional[List[str]] = Field(default=None, min_items=1)
    """
    The value of this keyword MUST be an array.  This array MUST have at
    least one element.  Elements of this array MUST be strings, and MUST
    be unique.
    
    An object instance is valid against this keyword if its property set
    contains all elements in this keyword's array value.
    """

    enum: Optional[List[Any]] = Field(default=None, min_items=1)
    """
    The value of this keyword MUST be an array.  This array SHOULD have
    at least one element.  Elements in the array SHOULD be unique.
    
    Elements in the array MAY be of any type, including null.
    
    An instance validates successfully against this keyword if its value
    is equal to one of the elements in this keyword's array value.
    """

    """
    The following properties are taken from the JSON Schema definition
    but their definitions were adjusted to the OpenAPI Specification.
    """

    type: Optional[str] = None
    """
    **From OpenAPI spec:
    Value MUST be a string. Multiple types via an array are not supported.**
    
    From JSON Schema:
    The value of this keyword MUST be either a string or an array.  If it
    is an array, elements of the array MUST be strings and MUST be
    unique.
    
    String values MUST be one of the seven primitive types defined by the
    core specification.
    
    An instance matches successfully if its primitive type is one of the
    types defined by keyword.  Recall: "number" includes "integer".
    """

    allOf: Optional[List[Union[Reference, "Schema"]]] = None
    """
    **From OpenAPI spec:
    Inline or referenced schema MUST be of a [Schema Object](#schemaObject) and not a standard JSON Schema.**
    
    From JSON Schema:
    This keyword's value MUST be an array.  This array MUST have at least
    one element.
    
    Elements of the array MUST be objects.  Each object MUST be a valid
    JSON Schema.
    
    An instance validates successfully against this keyword if it
    validates successfully against all schemas defined by this keyword's
    value.
    """

    oneOf: Optional[List[Union[Reference, "Schema"]]] = None
    """
    **From OpenAPI spec:
    Inline or referenced schema MUST be of a [Schema Object](#schemaObject) and not a standard JSON Schema.**
    
    From JSON Schema:
    This keyword's value MUST be an array.  This array MUST have at least
    one element.
    
    Elements of the array MUST be objects.  Each object MUST be a valid
    JSON Schema.
    
    An instance validates successfully against this keyword if it
    validates successfully against exactly one schema defined by this
    keyword's value.
    """

    anyOf: Optional[List[Union[Reference, "Schema"]]] = None
    """
    **From OpenAPI spec:
    Inline or referenced schema MUST be of a [Schema Object](#schemaObject) and not a standard JSON Schema.**
    
    From JSON Schema:
    This keyword's value MUST be an array.  This array MUST have at least
    one element.
    
    Elements of the array MUST be objects.  Each object MUST be a valid
    JSON Schema.
    
    An instance validates successfully against this keyword if it
    validates successfully against at least one schema defined by this
    keyword's value.
    """

    schema_not: Optional[Union[Reference, "Schema"]] = Field(default=None, alias="not")
    """
    **From OpenAPI spec:
    Inline or referenced schema MUST be of a [Schema Object](#schemaObject) and not a standard JSON Schema.**
    
    From JSON Schema:
    This keyword's value MUST be an object.  This object MUST be a valid
    JSON Schema.
    
    An instance is valid against this keyword if it fails to validate
    successfully against the schema defined by this keyword.
    """

    items: Optional[Union[Reference, "Schema"]] = None
    """
    **From OpenAPI spec:
    Value MUST be an object and not an array.
    Inline or referenced schema MUST be of a [Schema Object](#schemaObject) and not a standard JSON Schema.
    `items` MUST be present if the `type` is `array`.**
    
    From JSON Schema:
    The value of "items" MUST be either a schema or array of schemas.

    Successful validation of an array instance with regards to these two
    keywords is determined as follows:
    
    - if "items" is not present, or its value is an object, validation
      of the instance always succeeds, regardless of the value of
      "additionalItems";
    - if the value of "additionalItems" is boolean value true or an
      object, validation of the instance always succeeds;
    - if the value of "additionalItems" is boolean value false and the
      value of "items" is an array, the instance is valid if its size is
      less than, or equal to, the size of "items".
    """

    properties: Optional[Dict[str, Union[Reference, "Schema"]]] = None
    """
    **From OpenAPI spec:
    Property definitions MUST be a [Schema Object](#schemaObject)
    and not a standard JSON Schema (inline or referenced).**
    
    From JSON Schema:
    The value of "properties" MUST be an object.  Each value of this
    object MUST be an object, and each object MUST be a valid JSON
    Schema.
    
    If absent, it can be considered the same as an empty object.
    """

    additionalProperties: Optional[Union[bool, Reference, "Schema"]] = None
    """
    **From OpenAPI spec:
    Value can be boolean or object.
    Inline or referenced schema MUST be of a [Schema Object](#schemaObject) and not a standard JSON Schema.
    Consistent with JSON Schema, `additionalProperties` defaults to `true`.**
    
    From JSON Schema:
    The value of "additionalProperties" MUST be a boolean or a schema.

    If "additionalProperties" is absent, it may be considered present
    with an empty schema as a value.
    
    If "additionalProperties" is true, validation always succeeds.
    
    If "additionalProperties" is false, validation succeeds only if the
    instance is an object and all properties on the instance were covered
    by "properties" and/or "patternProperties".
    
    If "additionalProperties" is an object, validate the value as a
    schema to all of the properties that weren't validated by
    "properties" nor "patternProperties".
    """

    description: Optional[str] = None
    """
    **From OpenAPI spec:
    [CommonMark syntax](https://spec.commonmark.org/) MAY be used for rich text representation.**
    
    From JSON Schema:
    The value "description" MUST be a string.

    The description can be used to decorate a user interface with
    information about the data produced by this user interface.
    The description will provide explanation about the purpose of
    the instance described by this schema.
    """

    schema_format: Optional[str] = Field(default=None, alias="format")
    """
    **From OpenAPI spec:
    [Data Type Formats](#dataTypeFormat) for further details.
    While relying on JSON Schema's defined formats, the OAS offers a few additional predefined formats.**
    
    From JSON Schema:
    Structural validation alone may be insufficient to validate that an
    instance meets all the requirements of an application.  The "format"
    keyword is defined to allow interoperable semantic validation for a
    fixed subset of values which are accurately described by
    authoritative resources, be they RFCs or other external
    specifications.
    
    The value of this keyword is called a format attribute.  It MUST be a
    string.  A format attribute can generally only validate a given set
    of instance types.  If the type of the instance to validate is not in
    this set, validation for this format attribute and instance SHOULD
    succeed.
    """

    default: Optional[Any] = None
    """
    **From OpenAPI spec:
    The default value represents what would be assumed by the consumer of the input
    as the value of the schema if one is not provided.
    Unlike JSON Schema, the value MUST conform to the defined type for the Schema Object defined at the same level.
    For example, if `type` is `string`, then `default` can be `"foo"` but cannot be `1`.**
    
    From JSON Schema:
    There are no restrictions placed on the value of this keyword.
    
    This keyword can be used to supply a default JSON value associated
    with a particular schema.  It is RECOMMENDED that a default value be
    valid against the associated schema.
    
    This keyword MAY be used in root schemas, and in any subschemas.
    """

    """
    Other than the JSON Schema subset fields, the following fields MAY be used for further schema documentation:
    """

    nullable: Optional[bool] = None
    """
    A `true` value adds `"null"` to the allowed type specified by the `type` keyword,
    only if `type` is explicitly defined within the same Schema Object.
    Other Schema Object constraints retain their defined behavior,
    and therefore may disallow the use of `null` as a value.
    A `false` value leaves the specified or default `type` unmodified.
    The default value is `false`.
    """

    discriminator: Optional[Discriminator] = None
    """
    Adds support for polymorphism.
    The discriminator is an object name that is used to differentiate between other schemas
    which may satisfy the payload description.
    See [Composition and Inheritance](#schemaComposition) for more details.
    """

    readOnly: Optional[bool] = None
    """
    Relevant only for Schema `"properties"` definitions.
    Declares the property as "read only".
    This means that it MAY be sent as part of a response but SHOULD NOT be sent as part of the request.
    If the property is marked as `readOnly` being `true` and is in the `required` list,
    the `required` will take effect on the response only.
    A property MUST NOT be marked as both `readOnly` and `writeOnly` being `true`.
    Default value is `false`.
    """

    writeOnly: Optional[bool] = None
    """
    Relevant only for Schema `"properties"` definitions.
    Declares the property as "write only".
    Therefore, it MAY be sent as part of a request but SHOULD NOT be sent as part of the response.
    If the property is marked as `writeOnly` being `true` and is in the `required` list,
    the `required` will take effect on the request only.
    A property MUST NOT be marked as both `readOnly` and `writeOnly` being `true`.
    Default value is `false`.
    """

    xml: Optional[XML] = None
    """
    This MAY be used only on properties schemas.
    It has no effect on root schemas.
    Adds additional metadata to describe the XML representation of this property.
    """

    externalDocs: Optional[ExternalDocumentation] = None
    """
    Additional external documentation for this schema.
    """

    example: Optional[Any] = None
    """
    A free-form property to include an example of an instance for this schema.
    To represent examples that cannot be naturally represented in JSON or YAML,
    a string value can be used to contain the example with escaping where necessary.
    """

    deprecated: Optional[bool] = None
    """ 
    Specifies that a schema is deprecated and SHOULD be transitioned out of usage.
    Default value is `false`.
    """

    class Config:
        extra = Extra.ignore
        allow_population_by_field_name = True
        schema_extra = {
            "examples": [
                {"type": "string", "format": "email"},
                {
                    "type": "object",
                    "required": ["name"],
                    "properties": {
                        "name": {"type": "string"},
                        "address": {"$ref": "#/components/schemas/Address"},
                        "age": {"type": "integer", "format": "int32", "minimum": 0},
                    },
                },
                {"type": "object", "additionalProperties": {"type": "string"}},
                {"type": "object", "additionalProperties": {"$ref": "#/components/schemas/ComplexModel"}},
                {
                    "type": "object",
                    "properties": {"id": {"type": "integer", "format": "int64"}, "name": {"type": "string"}},
                    "required": ["name"],
                    "example": {"name": "Puma", "id": 1},
                },
                {
                    "type": "object",
                    "required": ["message", "code"],
                    "properties": {
                        "message": {"type": "string"},
                        "code": {"type": "integer", "minimum": 100, "maximum": 600},
                    },
                },
                {
                    "allOf": [
                        {"$ref": "#/components/schemas/ErrorModel"},
                        {"type": "object", "required": ["rootCause"], "properties": {"rootCause": {"type": "string"}}},
                    ]
                },
                {
                    "type": "object",
                    "discriminator": {"propertyName": "petType"},
                    "properties": {"name": {"type": "string"}, "petType": {"type": "string"}},
                    "required": ["name", "petType"],
                },
                {
                    "description": "A representation of a cat. "
                    "Note that `Cat` will be used as the discriminator value.",
                    "allOf": [
                        {"$ref": "#/components/schemas/Pet"},
                        {
                            "type": "object",
                            "properties": {
                                "huntingSkill": {
                                    "type": "string",
                                    "description": "The measured skill for hunting",
                                    "default": "lazy",
                                    "enum": ["clueless", "lazy", "adventurous", "aggressive"],
                                }
                            },
                            "required": ["huntingSkill"],
                        },
                    ],
                },
                {
                    "description": "A representation of a dog. "
                    "Note that `Dog` will be used as the discriminator value.",
                    "allOf": [
                        {"$ref": "#/components/schemas/Pet"},
                        {
                            "type": "object",
                            "properties": {
                                "packSize": {
                                    "type": "integer",
                                    "format": "int32",
                                    "description": "the size of the pack the dog is from",
                                    "default": 0,
                                    "minimum": 0,
                                }
                            },
                            "required": ["packSize"],
                        },
                    ],
                },
            ]
        }



import logging
from typing import Any, List, Set, Type, TypeVar

from pydantic import BaseModel
from pydantic.schema import schema

from . import Components, OpenAPI, Reference, Schema

logger = logging.getLogger(__name__)

PydanticType = TypeVar("PydanticType", bound=BaseModel)
ref_prefix = "#/components/schemas/"


class PydanticSchema(Schema):
    """Special `Schema` class to indicate a reference from pydantic class"""

    schema_class: Type[PydanticType] = ...
    """the class that is used for generate the schema"""


def construct_open_api_with_schema_class(
    open_api: OpenAPI,
    schema_classes: List[Type[PydanticType]] = None,
    scan_for_pydantic_schema_reference: bool = True,
    by_alias: bool = True,
) -> OpenAPI:
    """
    Construct a new OpenAPI object, with the use of pydantic classes to produce JSON schemas

    :param open_api: the base `OpenAPI` object
    :param schema_classes: pydanitic classes that their schema will be used "#/components/schemas" values
    :param scan_for_pydantic_schema_reference: flag to indicate if scanning for `PydanticSchemaReference` class
                                               is needed for "#/components/schemas" value updates
    :param by_alias: construct schema by alias (default is True)
    :return: new OpenAPI object with "#/components/schemas" values updated.
             If there is no update in "#/components/schemas" values, the original `open_api` will be returned.
    """
    new_open_api: OpenAPI = open_api.copy(deep=True)
    if scan_for_pydantic_schema_reference:
        extracted_schema_classes = _handle_pydantic_schema(new_open_api)
        if schema_classes:
            schema_classes = list({*schema_classes, *_handle_pydantic_schema(new_open_api)})
        else:
            schema_classes = extracted_schema_classes

    if not schema_classes:
        return open_api

    schema_classes.sort(key=lambda x: x.__name__)
    logger.debug(f"schema_classes{schema_classes}")

    # update new_open_api with new #/components/schemas
    schema_definitions = schema(schema_classes, by_alias=by_alias, ref_prefix=ref_prefix)
    if not new_open_api.components:
        new_open_api.components = Components()
    if new_open_api.components.schemas:
        for existing_key in new_open_api.components.schemas:
            if existing_key in schema_definitions.get("definitions"):
                logger.warning(
                    f'"{existing_key}" already exists in {ref_prefix}. '
                    f'The value of "{ref_prefix}{existing_key}" will be overwritten.'
                )
        new_open_api.components.schemas.update(
            {key: Schema.parse_obj(schema_dict) for key, schema_dict in schema_definitions.get("definitions").items()}
        )
    else:
        new_open_api.components.schemas = {
            key: Schema.parse_obj(schema_dict) for key, schema_dict in schema_definitions.get("definitions").items()
        }
    return new_open_api


def _handle_pydantic_schema(open_api: OpenAPI) -> List[Type[PydanticType]]:
    """
    This function traverses the `OpenAPI` object and

    1. Replaces the `PydanticSchema` object with `Reference` object, with correct ref value;
    2. Extracts the involved schema class from `PydanticSchema` object.

    **This function will mutate the input `OpenAPI` object.**

    :param open_api: the `OpenAPI` object to be traversed and mutated
    :return: a list of schema classes extracted from `PydanticSchema` objects
    """

    pydantic_types: Set[Type[PydanticType]] = set()

    def _traverse(obj: Any):
        if isinstance(obj, BaseModel):
            fields = obj.__fields_set__
            for field in fields:
                child_obj = obj.__getattribute__(field)
                if isinstance(child_obj, PydanticSchema):
                    logger.debug(f"PydanticSchema found in {obj.__repr_name__()}: {child_obj}")
                    obj.__setattr__(field, _construct_ref_obj(child_obj))
                    pydantic_types.add(child_obj.schema_class)
                else:
                    _traverse(child_obj)
        elif isinstance(obj, list):
            for index, elem in enumerate(obj):
                if isinstance(elem, PydanticSchema):
                    logger.debug(f"PydanticSchema found in list: {elem}")
                    obj[index] = _construct_ref_obj(elem)
                    pydantic_types.add(elem.schema_class)
                else:
                    _traverse(elem)
        elif isinstance(obj, dict):
            for key, value in obj.items():
                if isinstance(value, PydanticSchema):
                    logger.debug(f"PydanticSchema found in dict: {value}")
                    obj[key] = _construct_ref_obj(value)
                    pydantic_types.add(value.schema_class)
                else:
                    _traverse(value)

    _traverse(open_api)
    return list(pydantic_types)


def _construct_ref_obj(pydantic_schema: PydanticSchema):
    ref_obj = Reference(ref=ref_prefix + pydantic_schema.schema_class.__name__)
    logger.debug(f"ref_obj={ref_obj}")
    return ref_obj



from pydantic import BaseModel, Extra, Field


class Reference(BaseModel):
    """
    A simple object to allow referencing other components in the specification, internally and externally.

    The Reference Object is defined by [JSON Reference](https://tools.ietf.org/html/draft-pbryan-zyp-json-ref-03)
    and follows the same structure, behavior and rules.

    For this specification, reference resolution is accomplished as defined by the JSON Reference specification
    and not by the JSON Schema specification.
    """

    ref: str = Field(alias="$ref")
    """**REQUIRED**. The reference string."""

    class Config:
        extra = Extra.ignore
        allow_population_by_field_name = True
        schema_extra = {
            "examples": [{"$ref": "#/components/schemas/Pet"}, {"$ref": "Pet.json"}, {"$ref": "definitions.json#/Pet"}]
        }
